from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar, Union

from dagster import _check as check
from dagster._record import record
from pydantic import BaseModel

# Type variable for generic class handling
T = TypeVar("T")

# Constant for object attribute name
SCAFFOLDER_CLS_ATTRIBUTE = "__scaffolder_cls__"


def scaffold_with(
    scaffolder_cls: Union[type["Scaffolder"], "ScaffolderUnavailableReason"],
) -> Callable[[type[T]], type[T]]:
    """A decorator that declares what scaffolder is used to scaffold the artifact.

    Args:
        scaffolder_cls: A class that inherits from Scaffolder

    Returns:
        Decorator function that enhances the target class
    """

    def decorator(cls: type[T]) -> type[T]:
        # Store the scaffolder class as an attribute using the constant
        setattr(cls, SCAFFOLDER_CLS_ATTRIBUTE, scaffolder_cls)
        return cls

    return decorator


def has_scaffolder(cls: type) -> bool:
    """Determines if a class has been decorated with `@scaffold_with`.

    Args:
        cls: The class to check

    Returns:
        True if the class has a Scaffolder attached, False otherwise
    """
    return hasattr(cls, SCAFFOLDER_CLS_ATTRIBUTE)


def get_scaffolder(
    cls: type,
) -> Union["Scaffolder", "ScaffolderUnavailableReason"]:
    """Retrieves the scaffolder class attached to the decorated class.

    Args:
        cls: The class to inspect

    Returns:
        The scaffolder class attached to the decorated class. Raises CheckError if the class is not decorated with @scaffold_with.
    """
    check.param_invariant(has_scaffolder(cls), "cls", "Class must be decorated with @scaffold_with")
    attr = getattr(cls, SCAFFOLDER_CLS_ATTRIBUTE)
    return attr if isinstance(attr, ScaffolderUnavailableReason) else attr()


@dataclass
class ScaffolderUnavailableReason:
    message: str


@record
class ScaffoldRequest:
    # fully qualified class name of the decorated class
    type_name: str
    # target path for the scaffold request. Typically used to construct absolute paths
    target_path: Path


class Scaffolder:
    @classmethod
    def get_scaffold_params(cls) -> Optional[type[BaseModel]]:
        return None

    @abstractmethod
    def scaffold(self, request: ScaffoldRequest, params: Any) -> None: ...



from dagster_components.components.shim_components.base import ShimComponent, ShimScaffolder
from dagster_components.scaffold import scaffold_with


class AssetScaffolder(ShimScaffolder):
    def get_text(self) -> str:
        return """# import dagster as dg
# 
#
# @dg.asset
# def my_asset(context: dg.AssetExecutionContext) -> dg.MaterializeResult: ...
"""


@scaffold_with(AssetScaffolder)
class RawAssetComponent(ShimComponent):
    """Asset definition component."""



from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, Union

from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.result import MaterializeResult
from dagster_sling import DagsterSlingTranslator, SlingResource, sling_assets
from dagster_sling.resources import AssetExecutionContext
from pydantic import Field
from typing_extensions import TypeAlias

from dagster_components import Component, ComponentLoadContext
from dagster_components.components.sling_replication_collection.scaffolder import (
    SlingReplicationComponentScaffolder,
)
from dagster_components.resolved.context import ResolutionContext
from dagster_components.resolved.core_models import (
    AssetAttributesModel,
    AssetPostProcessor,
    AssetPostProcessorModel,
    OpSpec,
    OpSpecModel,
)
from dagster_components.resolved.metadata import ResolvableFieldInfo
from dagster_components.resolved.model import ResolvableModel, ResolvedFrom, Resolver
from dagster_components.scaffold import scaffold_with
from dagster_components.utils import TranslatorResolvingInfo

SlingMetadataAddons: TypeAlias = Literal["column_metadata", "row_count"]


class DagsterSlingTranslatorAutomationCondition(DagsterSlingTranslator):
    """Subclassed to pull in any overridden impl for get_automation_condition
    into produced asset specs.
    """

    def get_automation_condition(
        self, stream_definition: Mapping[str, Any]
    ) -> Optional[AutomationCondition]: ...

    def get_asset_spec(self, stream_definition: Mapping[str, Any]) -> AssetSpec:
        asset_spec = super().get_asset_spec(stream_definition)
        return asset_spec.replace_attributes(
            automation_condition=self.get_automation_condition(stream_definition)
        )


class ComponentsDagsterSlingTranslator(DagsterSlingTranslator):
    def __init__(self, *, resolving_info: TranslatorResolvingInfo):
        super().__init__()
        self.resolving_info = resolving_info

    def get_asset_spec(self, obj: Any) -> AssetSpec:
        base_spec = super().get_asset_spec(obj)
        return self.resolving_info.get_asset_spec(
            base_spec,
            {self.resolving_info.obj_name: obj, "spec": base_spec},
        )


def resolve_translator(
    context: ResolutionContext, model: "SlingReplicationModel"
) -> DagsterSlingTranslator:
    # TODO: Consider supporting owners and code_version in the future
    if model.asset_attributes and isinstance(model.asset_attributes, AssetAttributesModel):
        if model.asset_attributes.owners:
            raise ValueError("owners are not supported for sling_replication_collection component")
        if model.asset_attributes.code_version:
            raise ValueError(
                "code_version is not supported for sling_replication_collection component"
            )

    return ComponentsDagsterSlingTranslator(
        resolving_info=TranslatorResolvingInfo(
            "stream_definition",
            model.asset_attributes or AssetAttributesModel(),
            context,
        )
    )


@dataclass
class SlingReplicationSpecModel(ResolvedFrom["SlingReplicationModel"]):
    path: str
    op: Annotated[Optional[OpSpec], Resolver.from_annotation()]
    translator: Annotated[Optional[DagsterSlingTranslator], Resolver.from_model(resolve_translator)]
    include_metadata: list[SlingMetadataAddons]


class SlingReplicationModel(ResolvableModel):
    path: str = Field(
        ...,
        description="The path to the Sling replication file. For more information, see https://docs.slingdata.io/concepts/replication#overview.",
    )
    op: Optional[OpSpecModel] = Field(
        None,
        description="Customizations to the op underlying the Sling replication.",
    )
    include_metadata: list[SlingMetadataAddons] = Field(
        ["column_metadata", "row_count"],
        description="The metadata to include on materializations of the assets produced by the Sling replication.",
    )
    asset_attributes: Annotated[
        Optional[Union[str, AssetAttributesModel]],
        ResolvableFieldInfo(required_scope={"stream_definition"}),
    ] = Field(
        None,
        description="Customizations to the assets produced by the Sling replication.",
    )


class SlingReplicationCollectionModel(ResolvableModel):
    sling: Optional[SlingResource] = None
    replications: Sequence[SlingReplicationModel]
    asset_post_processors: Optional[Sequence[AssetPostProcessorModel]] = None


def resolve_resource(
    context: ResolutionContext, model: SlingReplicationCollectionModel
) -> SlingResource:
    return (
        SlingResource(**context.resolve_value(model.sling.model_dump()))
        if model.sling
        else SlingResource()
    )


@scaffold_with(SlingReplicationComponentScaffolder)
@dataclass
class SlingReplicationCollectionComponent(Component, ResolvedFrom[SlingReplicationCollectionModel]):
    """Expose one or more Sling replications to Dagster as assets.

    [Sling](https://slingdata.io/) is a Powerful Data Integration tool enabling seamless ELT
    operations as well as quality checks across files, databases, and storage systems.

    dg scaffold component dagster_components.dagster_sling.SlingReplicationCollectionComponent {component_name} to get started.

    This will create a component.yaml as well as a `replication.yaml` which is a Sling-specific configuration
    file. See Sling's [documentation](https://docs.slingdata.io/concepts/replication#overview) on `replication.yaml`.
    """

    resource: Annotated[SlingResource, Resolver.from_model(resolve_resource)] = ...
    replications: Annotated[Sequence[SlingReplicationSpecModel], Resolver.from_annotation()] = ...
    asset_post_processors: Annotated[
        Optional[Sequence[AssetPostProcessor]], Resolver.from_annotation()
    ] = None

    def build_asset(
        self, context: ComponentLoadContext, replication_spec_model: SlingReplicationSpecModel
    ) -> AssetsDefinition:
        op_spec = replication_spec_model.op or OpSpec()

        @sling_assets(
            name=op_spec.name or Path(replication_spec_model.path).stem,
            op_tags=op_spec.tags,
            replication_config=context.path / replication_spec_model.path,
            dagster_sling_translator=replication_spec_model.translator,
            backfill_policy=op_spec.backfill_policy,
        )
        def _asset(context: AssetExecutionContext):
            yield from self.execute(
                context=context, sling=self.resource, replication_spec_model=replication_spec_model
            )

        return _asset

    def execute(
        self,
        context: AssetExecutionContext,
        sling: SlingResource,
        replication_spec_model: SlingReplicationSpecModel,
    ) -> Iterator[Union[AssetMaterialization, MaterializeResult]]:
        iterator = sling.replicate(context=context)
        if "column_metadata" in replication_spec_model.include_metadata:
            iterator = iterator.fetch_column_metadata()
        if "row_count" in replication_spec_model.include_metadata:
            iterator = iterator.fetch_row_count()
        yield from iterator

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        defs = Definitions(
            assets=[self.build_asset(context, replication) for replication in self.replications],
        )
        for post_processor in self.asset_post_processors or []:
            defs = post_processor.fn(defs)
        return defs



from dagster_components.component_scaffolding import (
    scaffold_component_yaml as scaffold_component_yaml,
)
from dagster_components.core.component import (
    Component as Component,
    ComponentLoadContext as ComponentLoadContext,
    component as component,
)
from dagster_components.core.component_defs_builder import (
    build_component_defs as build_component_defs,
    load_defs as load_defs,
)
from dagster_components.core.component_scaffolder import (
    DefaultComponentScaffolder as DefaultComponentScaffolder,
)
from dagster_components.resolved.context import ResolutionContext as ResolutionContext
from dagster_components.resolved.core_models import (
    AssetAttributesModel as AssetAttributesModel,
    AssetPostProcessorModel as AssetPostProcessorModel,
    AssetSpecModel as AssetSpecModel,
    OpSpecModel as OpSpecModel,
)
from dagster_components.resolved.metadata import ResolvableFieldInfo as ResolvableFieldInfo
from dagster_components.resolved.model import (
    ResolvableModel as ResolvableModel,
    ResolvedFrom as ResolvedFrom,
    ResolvedKwargs as ResolvedKwargs,
    Resolver as Resolver,
)
from dagster_components.scaffold import (
    Scaffolder as Scaffolder,
    ScaffolderUnavailableReason as ScaffolderUnavailableReason,
    ScaffoldRequest as ScaffoldRequest,
)
from dagster_components.version import __version__ as __version__



from collections.abc import Mapping
from pathlib import Path
from typing import Any, Optional

import click
import yaml

from dagster_components.core.component import Component
from dagster_components.scaffold import ScaffolderUnavailableReason, ScaffoldRequest, get_scaffolder


class ComponentDumper(yaml.Dumper):
    def write_line_break(self) -> None:
        # add an extra line break between top-level keys
        if self.indent == 0:
            super().write_line_break()
        super().write_line_break()


def scaffold_component_yaml(
    request: ScaffoldRequest, attributes: Optional[Mapping[str, Any]]
) -> None:
    with open(request.target_path / "component.yaml", "w") as f:
        component_data = {"type": request.type_name, "attributes": attributes or {}}
        yaml.dump(
            component_data, f, Dumper=ComponentDumper, sort_keys=False, default_flow_style=False
        )
        f.writelines([""])


def scaffold_component_instance(
    path: Path,
    component_type: type[Component],
    component_type_name: str,
    scaffold_params: Mapping[str, Any],
) -> None:
    from dagster_components.components.shim_components.base import ShimComponent

    click.echo(f"Creating a Dagster component instance folder at {path}.")
    if not path.exists():
        path.mkdir()
    scaffolder = get_scaffolder(component_type)

    if isinstance(scaffolder, ScaffolderUnavailableReason):
        raise Exception(
            f"Component type {component_type_name} does not have a scaffolder. Reason: {scaffolder.message}."
        )

    scaffolder.scaffold(
        ScaffoldRequest(
            type_name=component_type_name,
            target_path=path,
        ),
        scaffold_params,
    )

    if not issubclass(component_type, ShimComponent):
        component_yaml_path = path / "component.yaml"
        if not component_yaml_path.exists():
            raise Exception(
                f"Currently all components require a component.yaml file. Please ensure your implementation of scaffold writes this file at {component_yaml_path}."
            )



from pathlib import Path
from typing import Optional

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.module_loaders.load_defs_from_module import (
    load_definitions_from_module,
)
from dagster._core.definitions.module_loaders.utils import find_objects_in_module_of_types
from pydantic import Field

from dagster_components import Component, ComponentLoadContext
from dagster_components.components.definitions_component.scaffolder import (
    DefinitionsComponentScaffolder,
)
from dagster_components.resolved.model import ResolvableModel
from dagster_components.scaffold import scaffold_with


@scaffold_with(DefinitionsComponentScaffolder)
class DefinitionsComponent(Component, ResolvableModel):
    """Wraps an arbitrary set of Dagster definitions."""

    definitions_path: Optional[str] = Field(
        None, description="Relative path to a file containing Dagster definitions."
    )

    def _build_defs_for_path(self, context: ComponentLoadContext, defs_path: Path) -> Definitions:
        defs_module = context.load_component_relative_python_module(defs_path)
        defs_attrs = list(find_objects_in_module_of_types(defs_module, Definitions))

        if defs_attrs and self.definitions_path is None:
            raise ValueError(
                f"Found a Definitions object in {defs_path}, which is not supported when implicitly loading definitions. "
                "You may need to add an explicit `component.yaml` file to specify a DefinitionsComponent."
            )
        elif len(defs_attrs) > 1:
            raise ValueError(
                f"Found multiple Definitions objects in {defs_path}. At most one Definitions object "
                "may be specified when using the DefinitionsComponent."
            )
        elif len(defs_attrs) == 1:
            return defs_attrs[0]
        else:
            return load_definitions_from_module(defs_module)

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        defs_paths = (
            [Path(self.definitions_path)]
            if self.definitions_path
            else list(context.path.rglob("*.py"))
        )

        return Definitions.merge(
            *(self._build_defs_for_path(context, defs_path) for defs_path in defs_paths)
        )



from dagster_components.components.shim_components.base import ShimComponent, ShimScaffolder
from dagster_components.scaffold import scaffold_with


class ScheduleScaffolder(ShimScaffolder):
    def get_text(self) -> str:
        return """# import dagster as dg
# 
#
# @dg.schedule(cron_schedule=..., target=...)
# def my_schedule(context: dg.ScheduleEvaluationContext): ...

"""


@scaffold_with(ScheduleScaffolder)
class RawScheduleComponent(ShimComponent):
    """Schedule component."""



import contextlib
import contextvars
import dataclasses
import importlib
import importlib.metadata
import inspect
import sys
import textwrap
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Optional, TypedDict, TypeVar, Union

from dagster import _check as check
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterError
from dagster._utils import pushd
from dagster._utils.cached_method import cached_method
from dagster._utils.source_position import SourcePositionTree
from typing_extensions import Self

from dagster_components.core.component_key import ComponentKey
from dagster_components.core.component_scaffolder import DefaultComponentScaffolder
from dagster_components.resolved.context import ResolutionContext
from dagster_components.resolved.model import ResolvableModel, ResolvedFrom, resolve_model
from dagster_components.scaffold import ScaffolderUnavailableReason, get_scaffolder, scaffold_with
from dagster_components.utils import format_error_message


class ComponentsEntryPointLoadError(DagsterError):
    pass


class ComponentDeclNode(ABC):
    @abstractmethod
    def load(self, context: "ComponentLoadContext") -> Sequence["Component"]: ...

    @abstractmethod
    def get_source_position_tree(self) -> Optional[SourcePositionTree]: ...


@scaffold_with(DefaultComponentScaffolder)
class Component(ABC):
    @classmethod
    def get_schema(cls) -> Optional[type["ResolvableModel"]]:
        from dagster_components.resolved.model import ResolvedFrom, get_model_type

        if issubclass(cls, ResolvableModel):
            return cls

        if issubclass(cls, ResolvedFrom):
            return get_model_type(cls)
        return None

    @classmethod
    def get_additional_scope(cls) -> Mapping[str, Any]:
        return {}

    @abstractmethod
    def build_defs(self, context: "ComponentLoadContext") -> Definitions: ...

    @classmethod
    def load(cls, attributes: Optional["ResolvableModel"], context: "ComponentLoadContext") -> Self:
        if issubclass(cls, ResolvableModel):
            # If the Component is a DSLSchema, the attributes in this case are an instance of itself
            assert isinstance(attributes, cls)
            return attributes

        elif issubclass(cls, ResolvedFrom):
            return (
                resolve_model(attributes, cls, context.resolution_context.at_path("attributes"))
                if attributes
                else cls()
            )
        else:
            # If the Component does not implement anything from Resolved, try to instantiate it without
            # argument.
            return cls()

    @classmethod
    def get_metadata(cls) -> "ComponentTypeInternalMetadata":
        docstring = cls.__doc__
        clean_docstring = _clean_docstring(docstring) if docstring else None

        scaffolder = get_scaffolder(cls)

        if isinstance(scaffolder, ScaffolderUnavailableReason):
            raise DagsterError(
                f"Component {cls.__name__} is not scaffoldable: {scaffolder.message}"
            )

        component_schema = cls.get_schema()
        scaffold_params = scaffolder.get_scaffold_params()
        return {
            "summary": clean_docstring.split("\n\n")[0] if clean_docstring else None,
            "description": clean_docstring if clean_docstring else None,
            "scaffold_params_schema": None
            if scaffold_params is None
            else scaffold_params.model_json_schema(),
            "component_schema": None
            if component_schema is None
            else component_schema.model_json_schema(),
        }

    @classmethod
    def get_description(cls) -> Optional[str]:
        return inspect.getdoc(cls)


def _clean_docstring(docstring: str) -> str:
    lines = docstring.strip().splitlines()
    first_line = lines[0]
    if len(lines) == 1:
        return first_line
    else:
        rest = textwrap.dedent("\n".join(lines[1:]))
        return f"{first_line}\n{rest}"


class ComponentTypeInternalMetadata(TypedDict):
    summary: Optional[str]
    description: Optional[str]
    scaffold_params_schema: Optional[Any]  # json schema
    component_schema: Optional[Any]  # json schema


class ComponentTypeMetadata(ComponentTypeInternalMetadata):
    name: str
    namespace: str


def get_entry_points_from_python_environment(group: str) -> Sequence[importlib.metadata.EntryPoint]:
    if sys.version_info >= (3, 10):
        return importlib.metadata.entry_points(group=group)
    else:
        return importlib.metadata.entry_points().get(group, [])


DG_LIBRARY_ENTRY_POINT_GROUP = "dagster_dg.library"


def load_component_type(component_key: ComponentKey) -> type[Component]:
    module_name, attr = component_key.namespace, component_key.name
    try:
        module = importlib.import_module(module_name)
        if not hasattr(module, attr):
            raise DagsterError(f"Module `{module_name}` has no attribute `{attr}`.")
        component_type = getattr(module, attr)
        if not issubclass(component_type, Component):
            raise DagsterError(
                f"Attribute `{attr}` in module `{module_name}` is not a subclass of `dagster_components.Component`."
            )
        return component_type
    except ModuleNotFoundError as e:
        raise DagsterError(f"Module `{module_name}` not found.") from e
    except ImportError as e:
        raise DagsterError(f"Error loading module `{module_name}`.") from e


def discover_entry_point_component_types() -> dict[ComponentKey, type[Component]]:
    """Discover component types registered in the Python environment via the
    `dagster_components` entry point group.

    `dagster-components` itself registers multiple component entry points. We call these
    "builtin" component libraries. The `dagster_components` entry point resolves to published
    component types and is loaded by default. Other entry points resolve to various sets of test
    component types. This method will only ever load one builtin component library.
    """
    component_types: dict[ComponentKey, type[Component]] = {}
    entry_points = get_entry_points_from_python_environment(DG_LIBRARY_ENTRY_POINT_GROUP)

    for entry_point in entry_points:
        try:
            root_module = entry_point.load()
        except Exception as e:
            raise ComponentsEntryPointLoadError(
                format_error_message(f"""
                    Error loading entry point `{entry_point.name}` in group `{DG_LIBRARY_ENTRY_POINT_GROUP}`.
                    Please fix the error or uninstall the package that defines this entry point.
                """)
            ) from e

        if not isinstance(root_module, ModuleType):
            raise DagsterError(
                f"Invalid entry point {entry_point.name} in group {DG_LIBRARY_ENTRY_POINT_GROUP}. "
                f"Value expected to be a module, got {root_module}."
            )
        for name, component_type in get_component_types_in_module(root_module):
            key = ComponentKey(name=name, namespace=entry_point.value)
            component_types[key] = component_type
    return component_types


def discover_component_types(modules: Sequence[str]) -> dict[ComponentKey, type[Component]]:
    component_types: dict[ComponentKey, type[Component]] = {}
    for extra_module in modules:
        for name, component_type in get_component_types_in_module(
            importlib.import_module(extra_module)
        ):
            key = ComponentKey(name=name, namespace=extra_module)
            component_types[key] = component_type
    return component_types


def get_component_types_in_module(
    module: ModuleType,
) -> Iterable[tuple[str, type[Component]]]:
    for attr in dir(module):
        value = getattr(module, attr)
        if (
            isinstance(value, type)
            and issubclass(value, Component)
            and not inspect.isabstract(value)
        ):
            yield attr, value


T = TypeVar("T")


@dataclass
class DefinitionsModuleCache:
    """Cache used when loading a code location's component hierarchy.
    Stores resources and a cache to ensure we don't load the same component multiple times.
    """

    resources: Mapping[str, object]

    def load_defs(self, module: ModuleType) -> Definitions:
        """Loads a set of Dagster definitions from a components Python module.

        Args:
            module (ModuleType): The Python module to load definitions from.

        Returns:
            Definitions: The set of Dagster definitions loaded from the module.
        """
        return self._load_defs_inner(module)

    @cached_method
    def _load_defs_inner(self, module: ModuleType) -> Definitions:
        from dagster_components.core.component_decl_builder import module_to_decl_node
        from dagster_components.core.component_defs_builder import defs_from_components

        decl_node = module_to_decl_node(module)
        if not decl_node:
            raise Exception(f"No component found at module {module}")

        context = ComponentLoadContext(
            module_name=module.__name__,
            decl_node=decl_node,
            resolution_context=ResolutionContext.default(decl_node.get_source_position_tree()),
            module_cache=self,
        )
        with use_component_load_context(context):
            components = decl_node.load(context)
            return defs_from_components(
                resources=self.resources, context=context, components=components
            )


@dataclass
class ComponentLoadContext:
    """Context for loading a single component."""

    module_name: str
    decl_node: Optional[ComponentDeclNode]
    resolution_context: ResolutionContext
    module_cache: DefinitionsModuleCache

    @staticmethod
    def current() -> "ComponentLoadContext":
        context = active_component_load_context.get()
        if context is None:
            raise DagsterError(
                "No active component load context, `ComponentLoadContext.current()` must be called inside of a component's `build_defs` method"
            )
        return context

    @staticmethod
    def for_test(
        *,
        resources: Optional[Mapping[str, object]] = None,
        decl_node: Optional[ComponentDeclNode] = None,
    ) -> "ComponentLoadContext":
        return ComponentLoadContext(
            module_name="test",
            decl_node=decl_node,
            resolution_context=ResolutionContext.default(
                decl_node.get_source_position_tree() if decl_node else None
            ),
            module_cache=DefinitionsModuleCache(resources=resources or {}),
        )

    @property
    def path(self) -> Path:
        from dagster_components.core.component_decl_builder import (
            ComponentFolder,
            ImplicitDefinitionsComponentDecl,
            PythonComponentDecl,
            YamlComponentDecl,
        )

        if not isinstance(
            self.decl_node,
            (
                YamlComponentDecl,
                PythonComponentDecl,
                ImplicitDefinitionsComponentDecl,
                ComponentFolder,
            ),
        ):
            check.failed(f"Unsupported decl_node type {type(self.decl_node)}")

        return self.decl_node.path

    def with_rendering_scope(self, rendering_scope: Mapping[str, Any]) -> "ComponentLoadContext":
        return dataclasses.replace(
            self,
            resolution_context=self.resolution_context.with_scope(**rendering_scope),
        )

    def for_decl_node(self, decl_node: ComponentDeclNode) -> "ComponentLoadContext":
        return dataclasses.replace(self, decl_node=decl_node)

    def normalize_component_type_str(self, type_str: str) -> str:
        return f"{self.module_name}{type_str}" if type_str.startswith(".") else type_str

    def load_defs(self, module: ModuleType) -> Definitions:
        """Builds the set of Dagster definitions for a component module.

        This is useful for resolving dependencies on other components.
        """
        return self.module_cache.load_defs(module)

    def load_component_relative_python_module(self, file_path: Path) -> ModuleType:
        """Load a python module relative to the component's context path. This is useful for loading code
        the resides within the component directory, loaded during `build_defs` method of a component.

        Example:
            .. code-block:: python

                def build_defs(self, context: ComponentLoadContext) -> Definitions:
                    return load_definitions_from_module(
                        context.load_component_relative_python_module(
                            Path(self.definitions_path) if self.definitions_path else Path("definitions.py")
                        )
                    )

        In a typical setup you end up with module names such as "a_project.components.my_component.my_python_file".

        This handles "__init__.py" files by ending the module name at the parent directory
        (e.g "a_project.components.my_component") if the file resides at "a_project/defs/my_component/__init__.py".

        This calls importlib.import_module with that module name, going through the python module import system.

        It is as if one typed "import a_project.components.my_component.my_python_file" in the python interpreter.
        """
        abs_file_path = file_path.absolute()
        with pushd(str(self.path)):
            abs_context_path = self.path.absolute()
            # Problematic
            # See https://linear.app/dagster-labs/issue/BUILD-736/highly-suspect-hardcoding-of-components-string-is-component-relative
            component_module_relative_path = abs_context_path.parts[
                abs_context_path.parts.index("defs") + 2 :
            ]
            component_module_name = ".".join([self.module_name, *component_module_relative_path])
            if abs_file_path.name != "__init__.py":
                component_module_name = f"{component_module_name}.{abs_file_path.stem}"

            return importlib.import_module(component_module_name)


active_component_load_context: contextvars.ContextVar[Union[ComponentLoadContext, None]] = (
    contextvars.ContextVar("active_component_load_context", default=None)
)


@contextlib.contextmanager
def use_component_load_context(component_load_context: ComponentLoadContext):
    token = active_component_load_context.set(component_load_context)
    try:
        yield
    finally:
        active_component_load_context.reset(token)


COMPONENT_LOADER_FN_ATTR = "__dagster_component_loader_fn"


T_Component = TypeVar("T_Component", bound=Component)


def component(
    fn: Callable[[ComponentLoadContext], T_Component],
) -> Callable[[ComponentLoadContext], T_Component]:
    setattr(fn, COMPONENT_LOADER_FN_ATTR, True)
    return fn


def is_component_loader(obj: Any) -> bool:
    return getattr(obj, COMPONENT_LOADER_FN_ATTR, False)



from typing import Any

from dagster_components.scaffold import Scaffolder, ScaffoldRequest


class DefaultComponentScaffolder(Scaffolder):
    def scaffold(self, request: ScaffoldRequest, params: Any) -> None:
        # This will be deleted once all components are converted to the new ComponentScaffolder API
        from dagster_components.component_scaffolding import scaffold_component_yaml

        scaffold_component_yaml(request, params.model_dump() if params else {})



from pathlib import Path
from typing import Optional

import click
from pydantic import TypeAdapter

from dagster_components.component_scaffolding import scaffold_component_instance
from dagster_components.core.component import load_component_type
from dagster_components.core.component_key import ComponentKey
from dagster_components.scaffold import ScaffolderUnavailableReason, get_scaffolder


@click.group(name="scaffold")
def scaffold_cli() -> None:
    """Commands for scaffolding Dagster components and related entities."""


@scaffold_cli.command(name="component")
@click.argument("component_type", type=str)
@click.argument("component_path", type=Path)
@click.option("--json-params", type=str, default=None)
def scaffold_component_command(
    component_type: str,
    component_path: Path,
    json_params: Optional[str],
) -> None:
    key = ComponentKey.from_typename(component_type)
    component_type_cls = load_component_type(key)

    if json_params:
        scaffolder = get_scaffolder(component_type_cls)
        if isinstance(scaffolder, ScaffolderUnavailableReason):
            raise Exception(
                f"Component type {component_type} does not have a scaffolder. Reason: {scaffolder.message}."
            )
        scaffold_params = TypeAdapter(scaffolder.get_scaffold_params()).validate_json(json_params)
    else:
        scaffold_params = {}

    scaffold_component_instance(
        component_path,
        component_type_cls,
        component_type,
        scaffold_params,
    )



import sys
import traceback
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from types import GenericAlias
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from dagster import _check as check
from pydantic import BaseModel, ConfigDict
from typing_extensions import TypeAlias

try:
    # this type only exists in python 3.10+
    from types import UnionType  # type: ignore
except ImportError:
    UnionType = Union

if TYPE_CHECKING:
    from dagster_components.resolved.context import ResolutionContext


class ResolvableModel(BaseModel):
    model_config = ConfigDict(extra="forbid")


TModel = TypeVar("TModel", bound=ResolvableModel)


def get_model_type(
    resolved_from_type: type["ResolvedFrom"],
) -> type[ResolvableModel]:
    """Returns the first generic type argument (TModel) of the ResolvedFrom subclass at runtime."""
    check.param_invariant(
        issubclass(resolved_from_type, ResolvedFrom),
        "resolvable_from_type",
    )
    check.param_invariant(
        hasattr(resolved_from_type, "__orig_bases__"),
        "resolvable_from_type",
    )
    for base in resolved_from_type.__orig_bases__:  # type: ignore
        # Check if this base originates from ResolvedFrom
        origin = getattr(base, "__origin__", None)
        if origin is ResolvedFrom:
            type_args = get_args(base)
            if not type_args:
                raise ValueError("ResolvedFrom base found but no generic type arguments present")
            return type_args[0]

    raise ValueError("No generic type arguments found in ResolvedFrom subclass")


def get_resolved_kwargs_target_type(resolved_kwargs_type: type["ResolvedKwargs"]):
    """Returns the generic type arguments (TModel, TObject) of a ResolvedKwargs subclass at runtime."""
    check.param_invariant(
        issubclass(resolved_kwargs_type, ResolvedKwargs),
        "resolved_kwargs_type",
    )
    check.param_invariant(
        hasattr(resolved_kwargs_type, "__orig_bases__"),
        "resolved_kwargs_type",
    )
    for base in resolved_kwargs_type.__orig_bases__:  # type: ignore
        # Check if this base originates from ResolvedFrom
        origin = getattr(base, "__origin__", None)
        if origin is ResolvedKwargs:
            type_args = get_args(base)
            if not type_args:
                raise ValueError("ResolvedKwargs base found but no generic type arguments present")
            if len(type_args) != 2:
                raise ValueError(
                    f"ResolvedKwargs base found but has incorrect number of type arguments, expected 2 got {len(type_args)}"
                )
            return type_args[1]

    raise ValueError("No generic type arguments found in ResolvedKwargs subclass")


T = TypeVar("T")


class ResolvedFrom(Generic[TModel], ABC):
    """Class which defines a type that can be resolved from the associated ResolvableModel."""


class ResolvedKwargs(Generic[TModel, T], ABC):
    """For cases where you can not inherit from ResolvedFrom on the desired target type,
    ResolvedKwargs allows you to define an object which will be resolved from its associated
    ResolvableModel and then passed unpacked as kwargs to create the target type.
    """


class _ModelResolver(Generic[T], ABC):
    @abstractmethod
    def resolve_from_model(self, context: "ResolutionContext", model: ResolvableModel) -> T: ...

    def from_seq(self, context: "ResolutionContext", model: Sequence[TModel]) -> Sequence[T]:
        return [self.resolve_from_model(context, item) for item in model]

    def from_optional(self, context: "ResolutionContext", model: Optional[TModel]) -> Optional[T]:
        return self.resolve_from_model(context, model) if model else None

    def from_optional_seq(
        self, context: "ResolutionContext", model: Optional[Sequence[TModel]]
    ) -> Optional[Sequence[T]]:
        return self.from_seq(context, model) if model else None


@dataclass(frozen=True)
class _KwargsResolver(_ModelResolver[T]):
    target_type: type[T]
    kwargs_type: type["ResolvedKwargs"]

    def resolve_from_model(self, context: "ResolutionContext", model: ResolvableModel) -> T:
        return resolve_model_using_kwargs_cls(
            model=model,
            kwargs_cls=self.kwargs_type,
            context=context,
            target_type=self.target_type,
        )


@dataclass(frozen=True)
class _DirectResolver(_ModelResolver[ResolvedFrom]):
    target_type: type[ResolvedFrom]

    def resolve_from_model(
        self, context: "ResolutionContext", model: ResolvableModel
    ) -> ResolvedFrom:
        return resolve_model(
            model=model,
            resolvable_type=self.target_type,
            context=context,
        )


@dataclass
class ParentFn:
    callable: Callable[["ResolutionContext", Any], Any]


@dataclass
class AttrWithContextFn:
    callable: Callable[["ResolutionContext", Any], Any]


@dataclass(frozen=True)
class _AutoResolve:
    via: Optional[type[ResolvedKwargs]] = None


class _ExpectedInjection: ...


def _is_scalar(annotation):
    if annotation in (int, float, str, bool, Any, type(None)):
        return True

    origin = get_origin(annotation)
    args = get_args(annotation)

    if origin in (Union, UnionType, list, Sequence, tuple, dict, Mapping) and all(
        _is_scalar(arg) for arg in args
    ):
        return True

    if origin is Annotated and any(isinstance(arg, _ExpectedInjection) for arg in args):
        return True

    if origin is Literal and all(_is_scalar(type(arg)) for arg in args):
        return True

    return False


def _get_resolver(
    annotation: type,
    top_level_auto_resolve: Optional[_AutoResolve],
) -> Optional[_ModelResolver]:
    if (
        top_level_auto_resolve
        and isinstance(annotation, type)
        and not isinstance(annotation, GenericAlias)  # prevent exceptions on 3.9
    ):
        if top_level_auto_resolve.via and annotation is get_resolved_kwargs_target_type(
            top_level_auto_resolve.via
        ):
            return _KwargsResolver(
                kwargs_type=top_level_auto_resolve.via,
                target_type=annotation,
            )
        if issubclass(annotation, ResolvedFrom):
            return _DirectResolver(annotation)

    origin = get_origin(annotation)
    if origin is not Annotated:
        return None

    args = get_args(annotation)
    auto_resolve = next((arg for arg in args if isinstance(arg, _AutoResolve)), None)
    if not auto_resolve:
        return None

    if auto_resolve.via:
        return _KwargsResolver(kwargs_type=auto_resolve.via, target_type=args[0])

    resolved_from_cls = args[0]
    if not issubclass(resolved_from_cls, ResolvedFrom):
        check.failed("Can only annotate ResolvedFrom types with ResolveModel()")
    return _DirectResolver(resolved_from_cls)


def derive_field_resolver(annotation: Any, field_name: str) -> "Resolver":
    if _is_scalar(annotation):
        return Resolver.from_model(
            lambda context, model: context.resolve_value(getattr(model, field_name))
        )

    origin = get_origin(annotation)
    args = get_args(annotation)

    top_level_auto_resolve = None
    if origin is Annotated:
        resolver = next((arg for arg in args if isinstance(arg, Resolver)), None)
        if resolver:
            return resolver

        top_level_auto_resolve = next((arg for arg in args if isinstance(arg, _AutoResolve)), None)
        resolver = _get_resolver(args[0], top_level_auto_resolve)
        if resolver:
            return Resolver(resolver.resolve_from_model)

        origin = get_origin(args[0])
        args = get_args(args[0])

    if origin in (Union, UnionType) and len(args) == 2:
        left_t, right_t = args
        if right_t is type(None):
            resolver = _get_resolver(left_t, top_level_auto_resolve)
            if resolver:
                return Resolver(resolver.from_optional)

            elif get_origin(left_t) in (Sequence, tuple, list):
                resolver = _get_resolver(get_args(left_t)[0], top_level_auto_resolve)
                if resolver:
                    return Resolver(resolver.from_optional_seq)

    elif origin in (Sequence, tuple, list):
        resolver = _get_resolver(args[0], top_level_auto_resolve)

        if resolver:
            return Resolver(resolver.from_seq)

    check.failed(
        f"Could not derive resolver for annotation {field_name}: {annotation}.\n"
        "Field types are expected to either be simple serializable types such as "
        "str, float, int, bool, list, etc or Annotated with an appropriate Resolver."
    )


class Resolver:
    """Contains information on how to resolve a field from a ResolvableModel."""

    def __init__(
        self,
        fn: Union[ParentFn, AttrWithContextFn, Callable[["ResolutionContext", Any], Any]],
    ):
        """Resolve this field by invoking the function which will receive the corresponding field value from the parent ResolvableModel."""
        if not isinstance(fn, (ParentFn, AttrWithContextFn)):
            if not callable(fn):
                check.param_invariant(
                    callable(fn),
                    "fn",
                    f"must be callable if not ParentFn or AttrWithContextFn. Got {fn}",
                )
            self.fn = AttrWithContextFn(fn)
        else:
            self.fn = fn
        super().__init__()

    @staticmethod
    def from_annotation():
        """Resolve this field to the (potentially nested) ResolvedFrom type via its associated ResolvableModel."""
        return _AutoResolve()

    @staticmethod
    def from_resolved_kwargs(kwargs_cls: type[ResolvedKwargs]):
        """Resolve this field to the (potentially nested) target type of the ResolvedKwargs via its associated ResolvableModel."""
        return _AutoResolve(via=kwargs_cls)

    @staticmethod
    def from_template_injection(via: Optional[type[ResolvedKwargs]] = None):
        """The complex type at this field must be template injected using an appropriate scope."""
        return _ExpectedInjection()

    @staticmethod
    def from_model(fn: Callable[["ResolutionContext", Any], Any]):
        """Resolve this field by invoking the function which will receive the entire parent ResolvableModel."""
        return Resolver(ParentFn(fn))

    def execute(
        self,
        context: "ResolutionContext",
        model: ResolvableModel,
        field_name: str,
    ) -> Any:
        from dagster_components.resolved.context import ResolutionException

        try:
            if isinstance(self.fn, ParentFn):
                return self.fn.callable(context, model)
            elif isinstance(self.fn, AttrWithContextFn):
                attr = getattr(model, field_name)
                return self.fn.callable(context.at_path(field_name), attr)
        except ResolutionException:
            raise  # already processed
        except Exception:
            raise context.build_resolve_fn_exc(
                traceback.format_exception(*sys.exc_info()),
                field_name=field_name,
                model=model,
            ) from None

        raise ValueError(f"Unsupported Resolver type: {self.fn}")


ResolvedType: TypeAlias = Union[type[ResolvedKwargs], type[ResolvedFrom]]


def get_annotation_field_resolvers(kwargs_cls: ResolvedType) -> dict[str, Resolver]:
    # Collect annotations from all base classes in MRO
    annotations = {}

    # Walk through all base classes in MRO
    for base in reversed(kwargs_cls.__mro__):
        # Get annotations from current base class if they exist
        base_annotations = getattr(base, "__annotations__", {})
        # Update annotations dict with any new annotations found
        # Later bases don't override earlier ones due to how update works
        annotations.update(base_annotations)

    return {
        field_name: derive_field_resolver(annotation, field_name)
        for field_name, annotation in annotations.items()
    }


def resolve_fields(
    model: ResolvableModel,
    kwargs_cls: Union[type[ResolvedFrom], type[ResolvedKwargs]],
    context: "ResolutionContext",
) -> Mapping[str, Any]:
    """Returns a mapping of field names to resolved values for those fields."""
    return {
        field_name: resolver.execute(
            context=context.at_path(field_name), model=model, field_name=field_name
        )
        for field_name, resolver in get_annotation_field_resolvers(kwargs_cls).items()
    }


TResolvedFrom = TypeVar("TResolvedFrom", bound=ResolvedFrom)


def resolve_model(
    model: ResolvableModel,
    resolvable_type: type[TResolvedFrom],
    context: "ResolutionContext",
) -> TResolvedFrom:
    return resolve_model_using_kwargs_cls(
        model=model,
        kwargs_cls=resolvable_type,
        context=context,
        target_type=resolvable_type,
    )


def resolve_model_using_kwargs_cls(
    model: ResolvableModel,
    kwargs_cls: Union[type[ResolvedFrom], type[ResolvedKwargs]],
    context: "ResolutionContext",
    target_type: type[T],
) -> T:
    # In the future we will do an explicit validation of alignment, but for now raise a marginally better error.
    check.inst_param(
        model,
        "model",
        ResolvableModel,
        "Ensure ResolveFrom field type annotations align with the corresponding ResolvableModel.",
    )
    return target_type(**resolve_fields(model, kwargs_cls, context))



import inspect
from collections.abc import Mapping, Sequence
from pathlib import Path
from types import ModuleType
from typing import Any, Optional, TypeVar, Union

from dagster import _check as check
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._record import record
from dagster._utils import pushd
from dagster._utils.pydantic_yaml import (
    _parse_and_populate_model_with_annotated_errors,
    enrich_validation_errors_with_source_position,
)
from dagster._utils.source_position import SourcePositionTree
from dagster._utils.yaml_utils import parse_yaml_with_source_positions
from pydantic import BaseModel, ConfigDict, TypeAdapter

from dagster_components.core.component import (
    Component,
    ComponentDeclNode,
    ComponentLoadContext,
    is_component_loader,
    load_component_type,
)
from dagster_components.core.component_key import ComponentKey
from dagster_components.utils import load_module_from_path


class ComponentFileModel(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: str
    attributes: Optional[Mapping[str, Any]] = None


T = TypeVar("T", bound=BaseModel)


def has_python_files(folder_path: Path) -> bool:
    """Check if a folder contains any Python files excluding __init__.py.

    Args:
        folder_path (Path): Path object representing the folder to check
    Returns:
        bool: True if folder contains .py files (except __init__.py), False otherwise
    """
    if not folder_path.is_dir():
        return False
    return any(folder_path.glob("*.py"))


@record
class ImplicitDefinitionsComponentDecl(ComponentDeclNode):
    path: Path

    @staticmethod
    def exists_at(path: Path) -> bool:
        if YamlComponentDecl.exists_at(path):
            return False
        if PythonComponentDecl.exists_at(path):
            return False
        return has_python_files(path)

    @staticmethod
    def from_path(path: Path) -> "ImplicitDefinitionsComponentDecl":
        return ImplicitDefinitionsComponentDecl(path=path)

    def get_source_position_tree(self) -> Optional[SourcePositionTree]:
        return None

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        from dagster_components.dagster import DefinitionsComponent

        return [DefinitionsComponent(definitions_path=None)]


@record
class PythonComponentDecl(ComponentDeclNode):
    path: Path

    @staticmethod
    def component_file_path(path: Path) -> Path:
        return path / "component.py"

    @staticmethod
    def exists_at(path: Path) -> bool:
        return PythonComponentDecl.component_file_path(path).exists()

    @staticmethod
    def from_path(path: Path) -> "PythonComponentDecl":
        return PythonComponentDecl(path=path)

    def get_source_position_tree(self) -> Optional[SourcePositionTree]:
        return None

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        module = load_module_from_path(
            self.path.stem, PythonComponentDecl.component_file_path(self.path)
        )
        component_loaders = list(inspect.getmembers(module, is_component_loader))
        if len(component_loaders) < 1:
            raise DagsterInvalidDefinitionError("No component loaders found in module")
        elif len(component_loaders) > 1:
            # note: we could support multiple component loaders in the same file, just
            # being more restrictive to start
            raise DagsterInvalidDefinitionError(
                f"Multiple component loaders found in module: {component_loaders}"
            )
        else:
            _, component_loader = component_loaders[0]
            return [component_loader(context)]


@record
class YamlComponentDecl(ComponentDeclNode):
    path: Path
    component_file_model: ComponentFileModel
    source_position_tree: Optional[SourcePositionTree] = None

    @staticmethod
    def component_file_path(path: Path) -> Path:
        return path / "component.yaml"

    @staticmethod
    def exists_at(path: Path) -> bool:
        return YamlComponentDecl.component_file_path(path).exists()

    @staticmethod
    def from_path(path: Path) -> "YamlComponentDecl":
        component_file_path = YamlComponentDecl.component_file_path(path)
        parsed = parse_yaml_with_source_positions(
            component_file_path.read_text(), str(component_file_path)
        )
        obj = _parse_and_populate_model_with_annotated_errors(
            cls=ComponentFileModel, obj_parse_root=parsed, obj_key_path_prefix=[]
        )

        return YamlComponentDecl(
            path=path,
            component_file_model=obj,
            source_position_tree=parsed.source_position_tree,
        )

    def get_source_position_tree(self) -> Optional[SourcePositionTree]:
        return self.source_position_tree

    def get_attributes(self, schema: type[T]) -> T:
        with pushd(str(self.path)):
            if self.source_position_tree:
                source_position_tree_of_attributes = self.source_position_tree.children[
                    "attributes"
                ]
                with enrich_validation_errors_with_source_position(
                    source_position_tree_of_attributes, ["attributes"]
                ):
                    return TypeAdapter(schema).validate_python(self.component_file_model.attributes)
            else:
                return TypeAdapter(schema).validate_python(self.component_file_model.attributes)

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        type_str = context.normalize_component_type_str(self.component_file_model.type)
        key = ComponentKey.from_typename(type_str)
        component_type = load_component_type(key)
        component_schema = component_type.get_schema()
        context = context.with_rendering_scope(component_type.get_additional_scope())

        attributes = self.get_attributes(component_schema) if component_schema else None
        return [component_type.load(attributes, context)]


@record
class ComponentFolder(ComponentDeclNode):
    path: Path
    sub_decls: Sequence[
        Union[
            YamlComponentDecl,
            PythonComponentDecl,
            ImplicitDefinitionsComponentDecl,
            "ComponentFolder",
        ]
    ]

    def load(self, context: ComponentLoadContext) -> Sequence[Component]:
        components = []
        for sub_decl in self.sub_decls:
            sub_context = context.for_decl_node(sub_decl)
            components.extend(sub_decl.load(sub_context))
        return components

    def get_source_position_tree(self) -> Optional[SourcePositionTree]:
        return None


def path_to_decl_node(path: Path) -> Optional[ComponentDeclNode]:
    # right now, we only support two types of components, both of which are folders
    # if the folder contains a component.yaml file, it's a component instance
    # otherwise, it's a folder containing sub-components

    if not path.is_dir():
        return None

    if YamlComponentDecl.exists_at(path):
        return YamlComponentDecl.from_path(path)
    elif PythonComponentDecl.exists_at(path):
        return PythonComponentDecl.from_path(path)
    elif ImplicitDefinitionsComponentDecl.exists_at(path):
        return ImplicitDefinitionsComponentDecl.from_path(path)

    subs = []
    for subpath in path.iterdir():
        component = path_to_decl_node(subpath)
        if component:
            subs.append(component)

    return ComponentFolder(path=path, sub_decls=subs)


def module_to_decl_node(module: ModuleType) -> Optional[ComponentDeclNode]:
    """Given a Python module, returns a corresponding component declaration node.

    Args:
        module (ModuleType): The Python module to convert to a component declaration node.

    Returns:
        Optional[ComponentDeclNode]: The corresponding component declaration node, or None if the module does not contain a component.
    """
    module_path = (
        Path(module.__file__).parent
        if module.__file__
        else Path(module.__path__[0])
        if module.__path__
        else None
    )
    return path_to_decl_node(
        check.not_none(module_path, f"Module {module.__name__} has no filepath")
    )



from collections.abc import Iterator, Mapping, Sequence
from dataclasses import dataclass, field
from functools import cached_property
from types import ModuleType
from typing import Annotated, Any, Optional, Union, cast

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster_dbt import (
    DagsterDbtTranslator,
    DbtCliResource,
    DbtManifestAssetSelection,
    DbtProject,
    dbt_assets,
)
from dagster_dbt.asset_utils import (
    get_asset_key_for_model as get_asset_key_for_model,
    get_asset_spec,
)
from dagster_dbt.dbt_manifest import validate_manifest
from dagster_dbt.utils import get_dbt_resource_props_by_dbt_unique_id_from_manifest
from typing_extensions import override

from dagster_components import Component, ComponentLoadContext
from dagster_components.components.dbt_project.scaffolder import DbtProjectComponentScaffolder
from dagster_components.resolved.core_models import (
    AssetAttributesModel,
    AssetPostProcessor,
    AssetPostProcessorModel,
    OpSpec,
    OpSpecModel,
    ResolutionContext,
)
from dagster_components.resolved.metadata import ResolvableFieldInfo
from dagster_components.resolved.model import ResolvableModel, ResolvedFrom, Resolver
from dagster_components.scaffold import scaffold_with
from dagster_components.utils import TranslatorResolvingInfo


class DbtProjectModel(ResolvableModel):
    dbt: DbtCliResource
    op: Optional[OpSpecModel] = None
    asset_attributes: Annotated[
        Optional[Union[str, AssetAttributesModel]],
        ResolvableFieldInfo(required_scope={"node"}),
    ] = None
    asset_post_processors: Optional[Sequence[AssetPostProcessorModel]] = None
    select: str = "fqn:*"
    exclude: Optional[str] = None


def resolve_translator(context: ResolutionContext, model: DbtProjectModel) -> DagsterDbtTranslator:
    class DagsterDbtTranslatorWithSpecs(DagsterDbtTranslator):
        def __init__(self, *, resolving_info: TranslatorResolvingInfo):
            super().__init__()
            self.resolving_info = resolving_info
            self.base_translator = DagsterDbtTranslator()
            self._specs_map: dict[int, AssetSpec] = {}

        @cached_property
        def project(self) -> DbtProject:
            return DbtProject(model.dbt.project_dir)

        @cached_property
        def manifest(self) -> Mapping[str, Any]:
            return validate_manifest(self.project.manifest_path)

        @cached_property
        def dbt_nodes(self) -> Mapping[str, Any]:
            return get_dbt_resource_props_by_dbt_unique_id_from_manifest(self.manifest)

        @cached_property
        def group_props(self) -> Mapping[str, Any]:
            return {group["name"]: group for group in self.manifest.get("groups", {}).values()}

        def get_asset_spec(self, stream_definition: Mapping[str, Any]) -> AssetSpec:
            if id(stream_definition) not in self._specs_map:
                base_spec = get_asset_spec(
                    translator=DagsterDbtTranslator(),
                    manifest=self.manifest,
                    dbt_nodes=self.dbt_nodes,
                    group_props=self.group_props,
                    project=self.project,
                    resource_props=stream_definition,
                )
                self._specs_map[id(stream_definition)] = self.resolving_info.get_asset_spec(
                    base_spec,
                    {self.resolving_info.obj_name: stream_definition, "spec": base_spec},
                )
            return self._specs_map[id(stream_definition)]

        @override
        def get_asset_key(self, dbt_resource_props: Mapping[str, Any]) -> AssetKey:
            return self.get_asset_spec(dbt_resource_props).key

        @override
        def get_description(self, dbt_resource_props: Mapping[str, Any]) -> Optional[str]:
            return self.get_asset_spec(dbt_resource_props).description

        @override
        def get_metadata(self, dbt_resource_props: Mapping[str, Any]) -> Mapping[str, Any]:
            return self.get_asset_spec(dbt_resource_props).metadata

        @override
        def get_tags(self, dbt_resource_props: Mapping[str, Any]) -> Mapping[str, str]:
            return self.get_asset_spec(dbt_resource_props).tags

        @override
        def get_group_name(self, dbt_resource_props: Mapping[str, Any]) -> Optional[str]:
            return self.get_asset_spec(dbt_resource_props).group_name

        @override
        def get_code_version(self, dbt_resource_props: Mapping[str, Any]) -> Optional[str]:
            return self.get_asset_spec(dbt_resource_props).code_version

        @override
        def get_owners(self, dbt_resource_props: Mapping[str, Any]) -> Sequence[str]:
            return self.get_asset_spec(dbt_resource_props).owners

        @override
        def get_automation_condition(
            self, dbt_resource_props: Mapping[str, Any]
        ) -> Optional[AutomationCondition]:
            return self.get_asset_spec(dbt_resource_props).automation_condition

    if (
        model.asset_attributes
        and isinstance(model.asset_attributes, AssetAttributesModel)
        and model.asset_attributes.deps
    ):
        # TODO: Consider supporting alerting deps in the future
        raise ValueError("deps are not supported for dbt_project component")
    return DagsterDbtTranslatorWithSpecs(
        resolving_info=TranslatorResolvingInfo(
            "node", model.asset_attributes or AssetAttributesModel(), context
        )
    )


def resolve_dbt(context: ResolutionContext, dbt: DbtCliResource) -> DbtCliResource:
    return DbtCliResource(**context.resolve_value(dbt.model_dump()))


@scaffold_with(DbtProjectComponentScaffolder)
@dataclass
class DbtProjectComponent(Component, ResolvedFrom[DbtProjectModel]):
    """Expose a DBT project to Dagster as a set of assets.

    This component assumes that you have already set up a dbt project. [Jaffle shop](https://github.com/dbt-labs/jaffle-shop) is their pre-existing
    example. Run `git clone --depth=1 https://github.com/dbt-labs/jaffle-shop.git jaffle_shop && rm -rf jaffle_shop/.git` to copy that project
    into your Dagster project directory.



    Scaffold by running `dagster scaffold component dagster_components.dagster_dbt.DbtProjectComponent --project-path path/to/your/existing/dbt_project`
    in the Dagster project directory.

    ### What is dbt?

    dbt is the industry standard for data transformation. Learn how it can help you transform
    data and deploy analytics code following software engineering best practices like
    version control, modularity, portability, CI/CD, and documentation.
    """

    dbt: Annotated[DbtCliResource, Resolver(resolve_dbt)]
    op: Annotated[Optional[OpSpec], Resolver.from_annotation()] = None
    # This requires from_parent because it access asset_attributes in the model
    translator: Annotated[DagsterDbtTranslator, Resolver.from_model(resolve_translator)] = field(
        default_factory=DagsterDbtTranslator
    )
    asset_post_processors: Annotated[
        Optional[Sequence[AssetPostProcessor]], Resolver.from_annotation()
    ] = None
    select: str = "fqn:*"
    exclude: Optional[str] = None

    @cached_property
    def project(self) -> DbtProject:
        return DbtProject(self.dbt.project_dir)

    def get_asset_selection(
        self, select: str, exclude: Optional[str] = None
    ) -> DbtManifestAssetSelection:
        return DbtManifestAssetSelection.build(
            manifest=self.project.manifest_path,
            dagster_dbt_translator=self.translator,
            select=select,
            exclude=exclude,
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        self.project.prepare_if_dev()

        @dbt_assets(
            manifest=self.project.manifest_path,
            project=self.project,
            name=self.op.name if self.op else self.project.name,
            op_tags=self.op.tags if self.op else None,
            dagster_dbt_translator=self.translator,
            select=self.select,
            exclude=self.exclude,
            backfill_policy=self.op.backfill_policy if self.op else None,
        )
        def _fn(context: AssetExecutionContext):
            yield from self.execute(context=context, dbt=self.dbt)

        defs = Definitions(assets=[_fn])
        for post_processor in self.asset_post_processors or []:
            defs = post_processor.fn(defs)
        return defs

    def execute(self, context: AssetExecutionContext, dbt: DbtCliResource) -> Iterator:
        yield from dbt.cli(["build"], context=context).stream()


def get_asset_key_for_model_from_module(
    context: ComponentLoadContext, dbt_component_module: ModuleType, model_name: str
) -> AssetKey:
    """Component-based version of dagster_dbt.get_asset_key_for_model. Returns the corresponding Dagster
    asset key for a dbt model, seed, or snapshot, loaded from the passed component path.

    Args:
        dbt_component_module (ModuleType): The module that was used to load the dbt project.
        model_name (str): The name of the dbt model, seed, or snapshot.

    Returns:
        AssetKey: The corresponding Dagster asset key.

    Examples:
        .. code-block:: python

            from dagster import asset
            from dagster_components.components.dbt_project import get_asset_key_for_model_from_module
            from dagster_components.core.component import ComponentLoadContext
            from my_project.defs import dbt_component

            ctx = ComponentLoadContext.get()

            @asset(deps={get_asset_key_for_model_from_module(ctx, dbt_component, "customers")})
            def cleaned_customers():
                ...
    """
    defs = context.load_defs(dbt_component_module)
    return get_asset_key_for_model(cast(Sequence[AssetsDefinition], defs.assets), model_name)



import importlib.util

_has_dagster_sling = importlib.util.find_spec("dagster_sling") is not None

if _has_dagster_sling:
    from dagster_components.components.sling_replication_collection.component import (
        SlingReplicationCollectionComponent as SlingReplicationCollectionComponent,
    )



import importlib.util

_has_dagster_dbt = importlib.util.find_spec("dagster_dbt") is not None

if _has_dagster_dbt:
    from dagster_components.components.dbt_project.component import (
        DbtProjectComponent as DbtProjectComponent,
    )



from dagster_components.components.shim_components.base import ShimComponent, ShimScaffolder
from dagster_components.scaffold import scaffold_with


class SensorScaffolder(ShimScaffolder):
    def get_text(self) -> str:
        return """# import dagster as dg
# 
#
# @dg.sensor(target=...)
# def my_sensor(context: dg.SensorEvaluationContext): ...
"""


@scaffold_with(SensorScaffolder)
class RawSensorComponent(ShimComponent):
    """Sensor component."""



import json
from typing import Any, Literal, Union

import click
from dagster._cli.utils import assert_no_remaining_opts
from dagster._cli.workspace.cli_target import (
    PythonPointerOpts,
    get_repository_python_origin_from_cli_opts,
    python_pointer_options,
)
from dagster._core.definitions.asset_job import is_reserved_asset_job_name
from dagster._utils.hosted_user_process import recon_repository_from_origin
from pydantic import ConfigDict, TypeAdapter, create_model

from dagster_components.core.component import (
    Component,
    ComponentTypeMetadata,
    discover_component_types,
    discover_entry_point_component_types,
)
from dagster_components.core.component_key import ComponentKey
from dagster_components.core.defs import (
    DgAssetMetadata,
    DgDefinitionMetadata,
    DgJobMetadata,
    DgScheduleMetadata,
    DgSensorMetadata,
)


@click.group(name="list")
def list_cli():
    """Commands for listing Dagster components and related entities."""


@list_cli.command(name="component-types")
@click.option("--entry-points/--no-entry-points", is_flag=True, default=True)
@click.argument("extra_modules", nargs=-1, type=str)
@click.pass_context
def list_component_types_command(
    ctx: click.Context, entry_points: bool, extra_modules: tuple[str, ...]
) -> None:
    """List registered Dagster components."""
    output: dict[str, Any] = {}
    component_types = _load_component_types(entry_points, extra_modules)
    for key in sorted(component_types.keys(), key=lambda k: k.to_typename()):
        output[key.to_typename()] = ComponentTypeMetadata(
            name=key.name,
            namespace=key.namespace,
            **component_types[key].get_metadata(),
        )
    click.echo(json.dumps(output))


@list_cli.command(name="all-components-schema")
@click.option("--entry-points/--no-entry-points", is_flag=True, default=True)
@click.argument("extra_modules", nargs=-1, type=str)
def list_all_components_schema_command(entry_points: bool, extra_modules: tuple[str, ...]) -> None:
    """Builds a JSON schema which ORs the schema for a component
    file for all component types available in the current code location.
    """
    component_types = _load_component_types(entry_points, extra_modules)

    schemas = []
    for key in sorted(component_types.keys(), key=lambda k: k.to_typename()):
        component_type = component_types[key]
        # Create ComponentFileModel schema for each type
        schema_type = component_type.get_schema()
        key_string = key.to_typename()
        if schema_type:
            schemas.append(
                create_model(
                    key.name,
                    type=(Literal[key_string], key_string),
                    attributes=(schema_type, None),
                    __config__=ConfigDict(extra="forbid"),
                )
            )
    union_type = Union[tuple(schemas)]  # type: ignore
    click.echo(json.dumps(TypeAdapter(union_type).json_schema()))


@list_cli.command(name="definitions")
@python_pointer_options
@click.pass_context
def list_definitions_command(ctx: click.Context, **other_opts: object) -> None:
    """List registered Dagster components."""
    python_pointer_opts = PythonPointerOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)

    repository_origin = get_repository_python_origin_from_cli_opts(python_pointer_opts)
    recon_repo = recon_repository_from_origin(repository_origin)
    repo_def = recon_repo.get_definition()

    all_defs: list[DgDefinitionMetadata] = []

    asset_graph = repo_def.asset_graph
    for key in sorted(list(asset_graph.get_all_asset_keys()), key=lambda key: key.to_user_string()):
        node = asset_graph.get(key)
        all_defs.append(
            DgAssetMetadata(
                type="asset",
                key=key.to_user_string(),
                deps=sorted([k.to_user_string() for k in node.parent_keys]),
                group=node.group_name,
                kinds=sorted(list(node.kinds)),
                description=node.description,
                automation_condition=node.automation_condition.__name__
                if node.automation_condition
                else None,
            )
        )
    for job in repo_def.get_all_jobs():
        if not is_reserved_asset_job_name(job.name):
            all_defs.append(DgJobMetadata(type="job", name=job.name))
    for schedule in repo_def.schedule_defs:
        schedule_str = (
            schedule.cron_schedule
            if isinstance(schedule.cron_schedule, str)
            else ", ".join(schedule.cron_schedule)
        )
        all_defs.append(
            DgScheduleMetadata(
                type="schedule",
                name=schedule.name,
                cron_schedule=schedule_str,
            )
        )
    for sensor in repo_def.sensor_defs:
        all_defs.append(DgSensorMetadata(type="sensor", name=sensor.name))

    click.echo(json.dumps(all_defs))


# ########################
# ##### HELPERS
# ########################


def _load_component_types(
    entry_points: bool, extra_modules: tuple[str, ...]
) -> dict[ComponentKey, type[Component]]:
    component_types = {}
    if entry_points:
        component_types.update(discover_entry_point_component_types())
    if extra_modules:
        component_types.update(discover_component_types(extra_modules))
    return component_types



from typing import Any

import yaml

from dagster_components.component_scaffolding import scaffold_component_yaml
from dagster_components.core.component_scaffolder import Scaffolder, ScaffoldRequest


class SlingReplicationComponentScaffolder(Scaffolder):
    def scaffold(self, request: ScaffoldRequest, params: Any) -> None:
        scaffold_component_yaml(request, {"replications": [{"path": "replication.yaml"}]})
        replication_path = request.target_path / "replication.yaml"
        with open(replication_path, "w") as f:
            yaml.dump({"source": {}, "target": {}, "streams": {}}, f)



from dagster_components.components.definitions_component.component import (
    DefinitionsComponent as DefinitionsComponent,
)
from dagster_components.components.pipes_subprocess_script_collection import (
    PipesSubprocessScriptCollectionComponent as PipesSubprocessScriptCollectionComponent,
)
from dagster_components.components.shim_components.asset import (
    RawAssetComponent as RawAssetComponent,
)
from dagster_components.components.shim_components.schedule import (
    RawScheduleComponent as RawScheduleComponent,
)
from dagster_components.components.shim_components.sensor import (
    RawSensorComponent as RawSensorComponent,
)



from abc import abstractmethod

from dagster_components import Component, Scaffolder, ScaffoldRequest


class ShimScaffolder(Scaffolder):
    @abstractmethod
    def get_text(self) -> str: ...
    def scaffold(self, request: ScaffoldRequest, params: None) -> None:
        if request.target_path.suffix != ".py":
            raise ValueError("Invalid target path suffix. Expected a path ending in `.py`.")
        # temporary hack as currently all scaffold requests target directories
        # that are auto-created
        request.target_path.rmdir()
        request.target_path.write_text(self.get_text())


class ShimComponent(Component):
    """A component that will never be loaded independently. This exists purely as a vessel
    for the @scaffoldable decorator so that the dagster-dg CLI can know about and invoke
    commands against it. In the near future, we'll improve the CLI such that it can handle
    arbitrary scaffoldable objects so that this hack is no longer necessary.
    """

    def build_defs(self, context):
        raise NotImplementedError("Shim components should never be loaded.")



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Annotated, Any, Callable, Literal, Optional, Union

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.asset_spec import AssetSpec, map_asset_specs
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._record import replace
from pydantic import BaseModel, Field
from typing_extensions import TypeAlias

from dagster_components.resolved.context import ResolutionContext
from dagster_components.resolved.model import (
    ResolvableModel,
    ResolvedFrom,
    ResolvedKwargs,
    Resolver,
    resolve_fields,
)


def _resolve_asset_key(key: str, context: ResolutionContext) -> AssetKey:
    resolved_val = context.resolve_value(key, as_type=AssetKey)
    return (
        AssetKey.from_user_string(resolved_val) if isinstance(resolved_val, str) else resolved_val
    )


PostProcessorFn: TypeAlias = Callable[[Definitions], Definitions]


class SingleRunBackfillPolicyModel(ResolvableModel):
    type: Literal["single_run"] = "single_run"


class MultiRunBackfillPolicyModel(ResolvableModel):
    type: Literal["multi_run"] = "multi_run"
    max_partitions_per_run: int = 1


def resolve_backfill_policy(
    context: ResolutionContext, model: "OpSpecModel"
) -> Optional[BackfillPolicy]:
    if model.backfill_policy is None:
        return None

    if model.backfill_policy.type == "single_run":
        return BackfillPolicy.single_run()
    elif model.backfill_policy.type == "multi_run":
        return BackfillPolicy.multi_run(
            max_partitions_per_run=model.backfill_policy.max_partitions_per_run
        )

    raise ValueError(f"Invalid backfill policy: {model.backfill_policy}")


@dataclass
class OpSpec(ResolvedFrom["OpSpecModel"]):
    name: Optional[str] = None
    tags: Optional[dict[str, str]] = None
    backfill_policy: Annotated[
        Optional[BackfillPolicy], Resolver.from_model(resolve_backfill_policy)
    ] = None


class OpSpecModel(ResolvableModel):
    name: Optional[str] = Field(default=None, description="The name of the op.")
    tags: Optional[dict[str, str]] = Field(
        default=None, description="Arbitrary metadata for the op."
    )
    backfill_policy: Optional[Union[SingleRunBackfillPolicyModel, MultiRunBackfillPolicyModel]] = (
        Field(default=None, description="The backfill policy to use for the assets.")
    )


class _ResolvableAssetAttributesMixin(BaseModel):
    deps: Sequence[str] = Field(
        default_factory=list,
        description="The asset keys for the upstream assets that this asset depends on.",
        examples=[["my_database/my_schema/upstream_table"]],
    )
    description: Optional[str] = Field(
        default=None,
        description="Human-readable description of the asset.",
        examples=["Refined sales data"],
    )
    metadata: Union[str, Mapping[str, Any]] = Field(
        default_factory=dict, description="Additional metadata for the asset."
    )
    group_name: Optional[str] = Field(
        default=None,
        description="Used to organize assets into groups, defaults to 'default'.",
        examples=["staging"],
    )
    skippable: bool = Field(
        default=False,
        description="Whether this asset can be omitted during materialization, causing downstream dependencies to skip.",
    )
    code_version: Optional[str] = Field(
        default=None,
        description="A version representing the code that produced the asset. Increment this value when the code changes.",
        examples=["3"],
    )
    owners: Sequence[str] = Field(
        default_factory=list,
        description="A list of strings representing owners of the asset. Each string can be a user's email address, or a team name prefixed with `team:`, e.g. `team:finops`.",
        examples=[["team:analytics", "nelson@hooli.com"]],
    )
    tags: Union[str, Mapping[str, str]] = Field(
        default_factory=dict,
        description="Tags for filtering and organizing.",
        examples=[{"tier": "prod", "team": "analytics"}],
    )
    kinds: Optional[Sequence[str]] = Field(
        default=None,
        description="A list of strings representing the kinds of the asset. These will be made visible in the Dagster UI.",
        examples=[["snowflake"]],
    )
    automation_condition: Optional[str] = Field(
        default=None,
        description="The condition under which the asset will be automatically materialized.",
    )


class AssetSpecModel(_ResolvableAssetAttributesMixin, ResolvableModel):
    key: str = Field(..., description="A unique identifier for the asset.")


class SharedAssetKwargs(ResolvedKwargs["AssetAttributesModel", AssetSpec]):
    deps: Sequence[str]
    description: Optional[str]
    metadata: Mapping[str, Any]
    group_name: Optional[str]
    skippable: bool
    code_version: Optional[str]
    owners: Sequence[str]
    tags: Mapping[str, str]
    kinds: Optional[Sequence[str]]
    automation_condition: Annotated[
        Optional[AutomationCondition], Resolver.from_template_injection()
    ]


class AssetSpecKwargs(SharedAssetKwargs):
    key: Annotated[
        AssetKey,
        Resolver.from_model(lambda context, schema: _resolve_asset_key(schema.key, context)),
    ]
    deps: Annotated[
        Sequence[AssetKey],
        Resolver.from_model(
            lambda context, schema: [_resolve_asset_key(dep, context) for dep in schema.deps]
        ),
    ]


class AssetAttributesKwargs(SharedAssetKwargs):
    key: Annotated[
        Optional[AssetKey],
        Resolver.from_model(
            lambda context, schema: _resolve_asset_key(schema.key, context) if schema.key else None
        ),
    ] = None


ResolvedAssetSpec: TypeAlias = Annotated[AssetSpec, Resolver.from_resolved_kwargs(AssetSpecKwargs)]


class AssetAttributesModel(_ResolvableAssetAttributesMixin, ResolvableModel):
    """Resolves into a dictionary of asset attributes. This is similar to AssetSpecSchema, but
    does not require a key. This is useful in contexts where you want to modify attributes of
    an existing AssetSpec.
    """

    key: Optional[str] = Field(default=None, description="A unique identifier for the asset.")


def resolve_asset_attributes_to_mapping(
    context: ResolutionContext,
    model: AssetAttributesModel,
) -> Mapping[str, Any]:
    # only include fields that are explcitly set
    set_fields = model.model_dump(exclude_unset=True).keys()
    resolved_fields = resolve_fields(model, AssetAttributesKwargs, context)
    return {k: v for k, v in resolved_fields.items() if k in set_fields}


ResolvedAssetAttributes: TypeAlias = Annotated[
    Mapping[str, Any], Resolver(resolve_asset_attributes_to_mapping)
]


class AssetPostProcessorModel(ResolvableModel):
    target: str = "*"
    operation: Literal["merge", "replace"] = "merge"
    attributes: AssetAttributesModel


def apply_post_processor_to_spec(
    model: AssetPostProcessorModel, spec: AssetSpec, context: ResolutionContext
) -> AssetSpec:
    # add the original spec to the context and resolve values
    attributes = resolve_asset_attributes_to_mapping(
        context=context.with_scope(asset=spec), model=model.attributes
    )

    if model.operation == "merge":
        mergeable_attributes = {"metadata", "tags"}
        merge_attributes = {k: v for k, v in attributes.items() if k in mergeable_attributes}
        replace_attributes = {k: v for k, v in attributes.items() if k not in mergeable_attributes}
        return spec.merge_attributes(**merge_attributes).replace_attributes(**replace_attributes)
    elif model.operation == "replace":
        return spec.replace_attributes(**attributes)
    else:
        check.failed(f"Unsupported operation: {model.operation}")


def apply_post_processor_to_defs(
    model: AssetPostProcessorModel, defs: Definitions, context: ResolutionContext
) -> Definitions:
    target_selection = AssetSelection.from_string(model.target, include_sources=True)
    target_keys = target_selection.resolve(defs.get_asset_graph())

    mappable = [d for d in defs.assets or [] if isinstance(d, (AssetsDefinition, AssetSpec))]
    mapped_assets = map_asset_specs(
        lambda spec: apply_post_processor_to_spec(model, spec, context)
        if spec.key in target_keys
        else spec,
        mappable,
    )

    assets = [
        *mapped_assets,
        *[d for d in defs.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))],
    ]
    return replace(defs, assets=assets)


def resolve_schema_to_post_processor(
    context, model: AssetPostProcessorModel
) -> Callable[[Definitions], Definitions]:
    return lambda defs: apply_post_processor_to_defs(model, defs, context)


@dataclass
class AssetPostProcessor(ResolvedFrom[AssetPostProcessorModel]):
    fn: Annotated[PostProcessorFn, Resolver.from_model(resolve_schema_to_post_processor)]



import importlib
from collections.abc import Mapping, Sequence
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Optional

from dagster import Definitions
from dagster._annotations import deprecated
from dagster._utils.warnings import suppress_dagster_warnings

from dagster_components.core.component import (
    Component,
    ComponentLoadContext,
    DefinitionsModuleCache,
    discover_entry_point_component_types,
)
from dagster_components.core.component_decl_builder import (
    ComponentDeclNode,
    ComponentFolder,
    YamlComponentDecl,
    path_to_decl_node,
)
from dagster_components.core.component_key import ComponentKey
from dagster_components.utils import get_path_from_module

if TYPE_CHECKING:
    from dagster import Definitions


def resolve_decl_node_to_yaml_decls(decl: ComponentDeclNode) -> list[YamlComponentDecl]:
    if isinstance(decl, YamlComponentDecl):
        return [decl]
    elif isinstance(decl, ComponentFolder):
        leaf_decls = []
        for sub_decl in decl.sub_decls:
            leaf_decls.extend(resolve_decl_node_to_yaml_decls(sub_decl))
        return leaf_decls

    raise NotImplementedError(f"Unknown component type {decl}")


def build_components_from_component_folder(
    context: ComponentLoadContext, path: Path
) -> Sequence[Component]:
    component_folder = path_to_decl_node(path)
    assert isinstance(component_folder, ComponentFolder)
    return component_folder.load(context.for_decl_node(component_folder))


@suppress_dagster_warnings
def defs_from_components(
    *,
    context: ComponentLoadContext,
    components: Sequence[Component],
    resources: Mapping[str, object],
) -> "Definitions":
    from dagster._core.definitions.definitions_class import Definitions

    return Definitions.merge(
        *[
            *[
                c.build_defs(context.with_rendering_scope(c.get_additional_scope()))
                for c in components
            ],
            Definitions(resources=resources),
        ]
    )


@deprecated(breaking_version="0.2.0")
@suppress_dagster_warnings
def build_component_defs(
    components_root: Path,
    resources: Optional[Mapping[str, object]] = None,
    component_types: Optional[dict[ComponentKey, type[Component]]] = None,
) -> "Definitions":
    """Build a Definitions object for all the component instances in a given code location.

    Args:
        components_root (Path): The path to the components root. This is a directory containing
            subdirectories with component instances.
    """
    defs_root = importlib.import_module(
        f"{Path(components_root).parent.name}.{Path(components_root).name}"
    )

    return load_defs(defs_root=defs_root, resources=resources, component_types=component_types)


# Public method so optional Nones are fine
@suppress_dagster_warnings
def load_defs(
    defs_root: ModuleType,
    resources: Optional[Mapping[str, object]] = None,
    component_types: Optional[dict[ComponentKey, type[Component]]] = None,
) -> "Definitions":
    """Constructs a Definitions object, loading all Dagster defs in the given module.

    Args:
        defs_root (Path): The path to the defs root, typically `package.defs`.
        resources (Optional[Mapping[str, object]]): A mapping of resource keys to resources
            to apply to the definitions.
        component_types (Optional[dict[ComponentKey, type[Component]]]): A mapping of
            component keys to component types.
    """
    from dagster._core.definitions.definitions_class import Definitions

    component_types = component_types or discover_entry_point_component_types()
    components_root_dir = get_path_from_module(defs_root)

    all_defs: list[Definitions] = []
    module_cache = DefinitionsModuleCache(resources=resources or {})
    for component_path in [item for item in components_root_dir.iterdir() if item.is_dir()]:
        defs = module_cache.load_defs(
            module=importlib.import_module(f"{defs_root.__name__}.{component_path.name}"),
        )
        all_defs.append(defs)
    return Definitions.merge(*all_defs)



from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from typing import TYPE_CHECKING, Annotated, Any, Callable, Generic, TypeVar, get_args, get_origin

from dagster._core.errors import DagsterInvalidDefinitionError
from pydantic import BaseModel, ConfigDict

from dagster_components.resolved.metadata import FieldInfo

if TYPE_CHECKING:
    from dagster_components.resolved.context import ResolutionContext


T = TypeVar("T")


def resolve_as(schema: "ResolvableSchema", target_type: type[T], context: "ResolutionContext") -> T:
    return target_type(**resolve_fields(schema, target_type, context))


def resolve_fields(
    schema: "ResolvableSchema", target_type: type, context: "ResolutionContext"
) -> Mapping[str, Any]:
    """Returns a mapping of field names to resolved values for those fields."""
    return {
        field_name: resolver.fn(context.at_path(field_name), schema)
        for field_name, resolver in get_field_resolvers(schema, target_type).items()
    }


def get_field_resolvers(
    schema: "ResolvableSchema", target_type: type
) -> Mapping[str, "FieldResolver"]:
    return {
        # extract field resolvers from annotations if possible, otherwise extract from the schema type
        **(
            _get_annotation_field_resolvers(target_type)
            or _get_annotation_field_resolvers(schema.__class__)
        )
    }


def get_resolved_type(schema: "ResolvableSchema") -> type:
    generic_base = next(
        base for base in schema.__class__.__bases__ if issubclass(base, ResolvableSchema)
    )
    generic_args = generic_base.__pydantic_generic_metadata__["args"]
    if len(generic_args) == 0:
        # if no generic type is specified, resolve back to the base type
        return schema.__class__
    elif len(generic_args) > 1:
        raise DagsterInvalidDefinitionError(
            f"Expected at most one generic argument for type: `{schema.__class__}`"
        )
    resolved_type = next(iter(generic_args))
    resolved_type = resolved_type if isinstance(resolved_type, type) else None
    if resolved_type is None:
        raise DagsterInvalidDefinitionError(
            f"Could not extract resolved type instance from `{schema.__class__}`. "
            "This can happen when using a ForwardRef when defining your ResolvableModel "
            '(`ResolvableModel["SomeType"]`). Consider using a concrete type or calling '
            "`resolve_as` instead."
        )
    return resolved_type


def resolve(schema: "ResolvableSchema", context: "ResolutionContext") -> object:
    return schema.resolve(context)


class ResolvableSchema(BaseModel, Generic[T]):
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    def resolve(self, context: "ResolutionContext") -> T:
        return resolve_as(self, get_resolved_type(self), context)


class FieldResolver(FieldInfo):
    """Contains information on how to resolve this field from a ResolvableSchema."""

    def __init__(self, fn: Callable[["ResolutionContext", Any], Any]):
        self.fn = fn
        super().__init__()

    @staticmethod
    def from_annotation(annotation: Any, field_name: str) -> "FieldResolver":
        if get_origin(annotation) is Annotated:
            args = get_args(annotation)
            resolver = next((arg for arg in args if isinstance(arg, FieldResolver)), None)
            if resolver:
                return resolver
        return FieldResolver(
            lambda context, schema: context.resolve_value(getattr(schema, field_name))
        )


def _get_annotation_field_resolvers(cls: type) -> dict[str, FieldResolver]:
    if issubclass(cls, BaseModel):
        # neither pydantic's Field.annotation nor Field.rebuild_annotation() actually
        # return the original annotation, so we have to walk the mro to get them
        annotations = {}
        for field_name in cls.model_fields:
            for base in cls.__mro__:
                if field_name in getattr(base, "__annotations__", {}):
                    annotations[field_name] = base.__annotations__[field_name]
                    break
        return {
            field_name: FieldResolver.from_annotation(annotation, field_name)
            for field_name, annotation in annotations.items()
        }
    elif is_dataclass(cls):
        return {
            field.name: FieldResolver.from_annotation(field.type, field.name)
            for field in fields(cls)
        }
    else:
        return {}
